import java.util.ArrayList;
import java.util.List;

import smile.Network;

public class Ego implements Cloneable 
{
	private int id;
	private double socEco;
	private int socEcoGroup;
	private int loc;	
	private int ob;	
	private List<Ego> fixedNeighbourList;
	private List<Ego> nearstNeighbourList;
	private double x;
	private double y;
	
	private double[] infHAdvProbs;
	private double[] hAdv_ReactProbs;
	private double[] infAdProbs;
	private double[] ad_ReactProbs;
	private double[] reactProbs;
	private double[] actionProbs;
		
	private List<Meme> memes;
	
	Ego(int id)
	{
		this.id = id;
		memes = new ArrayList<Meme>();
		infHAdvProbs = new double[3];
		hAdv_ReactProbs = new double[3];
		infAdProbs = new double[3];
		ad_ReactProbs = new double[3];
		reactProbs = new double[3];
		actionProbs = new double[3];
		fixedNeighbourList = new ArrayList<Ego>();
		nearstNeighbourList = new ArrayList<Ego>();
		x = Math.random() * 100;
		y = Math.random() * 100;
	}
	
	public int getLoc()
	{
		return loc;
	}
	
	public int getOb()
	{
		return ob;
	}
	public int getSocEco() throws Exception
	{
		if(socEco >= Sociaty.SocEco_rich_thr)
		{
			return 2;
		}
		else if(socEco >= Sociaty.SocEco_medium_thr)
		{
			return 1;
		}
		else if(socEco >=0)
		{
			return 0;
		}
		else
		{
			throw new Exception();
		}
	}
	public void addFixedNeighbour(Ego ego)
	{
		if(!fixedNeighbourList.contains(ego))
		{
			fixedNeighbourList.add(ego);
		}		
	}
	
	public void assignNearestNeighbours()
	{
		for (int i = 0; i < Sociaty.egos.size(); i++)
		{
			if(id != Sociaty.egos.get(i).getID())
			{
				if(calculateDistance(Sociaty.egos.get(i)) < Sociaty.neighbourRangeR)
				{
					nearstNeighbourList.add(Sociaty.egos.get(i));
				}
			}
			
		}
	}
	
	private double calculateDistance(Ego ego)
	{
		return Math.sqrt((x - ego.getX()) * (x - ego.getX()) +
				(y - ego.getY()) * (y - ego.getY()));
	}
	
	public double getX()
	{
		return x;
	}
	
	public double getY()
	{
		return y;
	}
	
	public int getID()
	{
		return id;
	}
	
	public void initialise(double socEco, int loc, int ob)
	{
		this.socEco = Math.max(socEco, 0);
		this.loc = loc;
		this.ob = ob;
	}
	
	public void addMeme(Meme meme)
	{
		int memeIndex = getMemeIndex(meme);
		if(memeIndex < 0)
		{
			//this meme does exist; generate a new meme
			memes.add(new Meme(meme.getSourceMeme(), meme.getEnery(), socEco));
		}		
		else
		{
			memes.get(memeIndex).merge(meme.getEnery(), socEco);			
		}
	}
	
	private int getMemeIndex(Meme meme)
	{
		int id = meme.getID();
		for(int i = 0; i < memes.size(); i++)
		{
			if(memes.get(i).getID() == id)
			{
				return i;
			}
		}
		return -1;
	}
	
	public void infoDefuse(int currentTime)
	{
		if(memes.isEmpty())
		{
			return;
		}
		else
		{
			for(int i = 0; i < memes.size(); i++)
			{
				int dufusionType = memes.get(i).getDefusionType();
				if(dufusionType == SourceMeme.fixedNetMode)
				{
					infoDefuseFixNets(currentTime, memes.get(i));
				}
				if(dufusionType == SourceMeme.randonNetMode)
				{
					infoDefuseRandomly(currentTime, memes.get(i));
				}
				if(dufusionType == SourceMeme.nearestNeibourghNetMode)
				{
					infoDefuseNearestNeighbour(currentTime, memes.get(i));	
				}
			}

		}
	}
	
	private void infoDefuseFixNets(int currentTime, Meme meme)
	{
		if(currentTime > meme.getHappeningTime())
		{
			for(int i = 0; i < fixedNeighbourList.size(); i++)
			{
				fixedNeighbourList.get(i).addMeme(meme);
			}
		}
	}

	private void infoDefuseNearestNeighbour(int currentTime, Meme meme)
	{
		if(currentTime > meme.getHappeningTime())
		{
			for(int i = 0; i < nearstNeighbourList.size(); i++)
			{
				nearstNeighbourList.get(i).addMeme(meme);
			}
		}
	}
	
	private void infoDefuseRandomly(int currentTime, Meme meme)
	{
		if(currentTime > meme.getHappeningTime())
		{
			int propagationPopulation = meme.getPropagationNum();
			for(int j = 0; j < propagationPopulation; j++)
			{
				int index = (int)Math.floor(Math.random() * Sociaty.population);
				Sociaty.egos.get(index).addMeme(meme);
			}
		}
	}
	
	public void calculateReaction() throws Exception
	{
		Sociaty.bayesNet.clearAllEvidence();
		setSocEoc(Sociaty.bayesNet);
		setLoc(Sociaty.bayesNet);	
		setOb(Sociaty.bayesNet);
		for(int i = 0; i < memes.size(); i++)
		{	
			setMeme(Sociaty.bayesNet, memes.get(i));
		}
		Sociaty.bayesNet.updateBeliefs();	
		setChildNodesProbs(Sociaty.bayesNet);
	}
	
	
	private void setSocEoc(Network bn) throws Exception
	{
		if(socEco >= Sociaty.SocEco_rich_thr)
		{
			bn.setEvidence("Soc_Eco", "rich");
		}
		else if(socEco >= Sociaty.SocEco_medium_thr)
		{
			bn.setEvidence("Soc_Eco", "medium");
		}
		else if(socEco >=0)
		{
			bn.setEvidence("Soc_Eco", "poor");
		}
		else
		{
			throw new Exception();
		}
	}
	private void setLoc(Network bn)
	{
		if(loc == Sociaty.loc_low)
		{
			bn.setEvidence("Loc", "low");
		}
		else if(loc == Sociaty.loc_medium)
		{
			bn.setEvidence("Loc", "medium");
		}
		else if(loc == Sociaty.loc_high)
		{
			bn.setEvidence("Loc", "high");
		}
		else
		{
			
		}
	}

	private void setOb(Network bn)
	{
		if(ob == Sociaty.ob_low)
		{
			bn.setEvidence("Ob", "low");
		}
		else if(ob == Sociaty.ob_medium)
		{
			bn.setEvidence("Ob", "medium");
		}
		else if(ob == Sociaty.ob_high)
		{
			bn.setEvidence("Ob", "high");
		}
		else
		{
			
		}
	}
	

	private void setMeme(Network bn, Meme meme)
	{
		double en = meme.getEnery();
		if(meme.isHADV())
		{
			if(en > 0.66)
			{
				bn.setEvidence("HAdv", "high");
			}
			else if(en > 0.33)
			{
				bn.setEvidence("HAdv", "medium");
			}
			else
			{
				bn.setEvidence("HAdv", "low");
			}
			
		}
		else
		{
			if(en > 0.66)
			{
				bn.setEvidence("Ad", "high");
			}
			else if(en > 0.33)
			{
				bn.setEvidence("Ad", "medium");
			}
			else
			{
				bn.setEvidence("Ad", "low");
			}
		}		
	}
	private void setChildNodesProbs(Network bn)
	{
		String[] nodeNames = bn.getAllNodeIds();
		for(int i = 0; i < nodeNames.length; i++)
		{
			String nodeName = nodeNames[i];
            double[] probs = bn.getNodeValue(nodeName);
               		
			if(nodeName.equals("InfHAdv"))
			{
				for(int j = 0; j < 3; j++)
				{
					infHAdvProbs[j] = probs[j];
				}
			}
			else if(nodeName.endsWith("HAdv_React"))
			{
				for(int j = 0; j < 3; j++)
				{
					hAdv_ReactProbs[j] = probs[j];
				}
				
			}
			else if(nodeName.endsWith("InfAd"))
			{
				for(int j = 0; j < 3; j++)
				{
					infAdProbs[j] = probs[j];
				}
			}
			else if(nodeName.endsWith("Ad_React"))
			{
				for(int j = 0; j < 3; j++)
				{
					ad_ReactProbs[j] = probs[j];
				}
			}
			else if(nodeName.endsWith("React"))
			{
				for(int j = 0; j < 3; j++)
				{
					reactProbs[j] = probs[j];
				}
			}
			else if(nodeName.endsWith("Action"))
			{
				for(int j = 0; j < 3; j++)
				{
					actionProbs[j] = probs[j];
				}
			}
			else 
			{
				
			}
			
		}		
	}	

	public void wane()
	{
		for(int i = 0; i < memes.size(); i++)
		{
			memes.get(i).wane();
		}
	}
	
	public void debugPrint()
	{
		System.out.println(id
				+ "," + getString0(socEco)
				+ "," + getString(loc)
				+ "," + getString(ob)
				+ "," + actionProbs[0]
				+ "," + actionProbs[1]
				+ "," + actionProbs[2]);
	}
	
	private String getString(int i)	
	{
		if(i == 0)
		{
			return "low";
		}
		else if(i == 1)
		{
			return "medium";
		}
		else if(i == 2)
		{
			return "high";
		}
		else
		{
			return "invalid";
		}
	}

	private String getString0(double var)	
	{
		if(var >= Sociaty.SocEco_rich_thr)
		{
			return "rich";
		}
		else if(var >= Sociaty.SocEco_medium_thr)
		{
			return "medium";
		}
		else if(var >= 0)
		{
			return "poor";
		}
		else
		{
			return "invalid";
		}
	}
	 public double getHAdvEnery()
	 {
		 double energy = 0.0d;
		 for(int i = 0; i < memes.size(); i++)
		 {
			 if(memes.get(i).isHADV())
			 {
				 energy += memes.get(i).getEnery();
			 }
		 }
		 return energy;
	 }

	 public double getAdEnery()
	 {
		 double energy = 0.0d;
		 for(int i = 0; i < memes.size(); i++)
		 {
			 if(memes.get(i).isAD())
			 {
				 energy += memes.get(i).getEnery();
			 }
		 }
		 return energy;
	 }
	 
	 public double[] getActionProbs()
	 {
		 return actionProbs;		 
	 }

	 public double[] getReactProbs()
	 {
		 return reactProbs;		 
	 }
}
	
